import json
from re import T
from typing import Optional
import sqlite3
from .Column import Column
from .Table import Table
from .Tools import *

class Database():
    def __init__(self, Path="", Password="", AutoSave=False):
        self.Tables = {}
        self.password = Password
        self.path = Path
        self.autosave = AutoSave
        if self.path != "":
            try:
                self.Load()
            except FileNotFoundError:
                raise FileNotFoundError(f"File '{self.path}' not found.")
            except Exception as e:
                raise e
    
    def __getitem__(self, Key) -> Table:
        return self.Tables[Key]
    
    def __getattr__(self, name: str) -> Optional[Table]:
        if name in self.Columns:
            return self.Columns[name]
        raise AttributeError(f"Table '{name}' not found.")
    
    def GetTables(self):
        return list(self.Tables.keys())
    
    def CreateTable(self, TableName) -> Table:
        self.Tables[TableName] = Table(TableName,self.SaveIfPossible)
        setattr(self.__class__, TableName, property(lambda self: self.Tables[TableName]))
        self.SaveIfPossible()
        return self.Tables[TableName]
    
    
    # @NotImplemented
    # def CreateTableForClass(self, TableName, Class) -> Table:
    #     self.Tables[TableName] = Table(TableName, self.autosave_callback if self.autosave else None)
    #     self.Tables[TableName].AddColumn(**Class.__dict__)
    #     setattr(Class, TableName, property(lambda self: self.Tables[TableName]))
    #     self.SaveIfPossible()
    #     return self.Tables[TableName]
    
    def RenameTable(self,OldName,NewName):
        self.Tables[NewName] = self.Tables.pop(OldName)
        self.Tables[NewName].TableName = NewName
        setattr(self.__class__, NewName, property(lambda self, name=NewName: self.Tables[name]))
        delattr(self.__class__, OldName)
        self.SaveIfPossible()
    
    def DeleteTable(self,TableName):
        del self.Tables[TableName]
        delattr(self.__class__, TableName)
        
    def __len__(self) -> int:
        return len(self.Tables.keys())
    
    def Load(self):
        wasOnAutoSave = self.autosave
        self.autosave = False
        fromJson = False
        if self.password == "" : fromJson = True
        try:
            with open(self.path,"r") as f:
                if fromJson:
                    data = json.loads(f.read())
                else:
                    data = json.loads(decrypt(f.read(),self.password))
                for k,v in data["Tables"].items():
                    t = self.CreateTable(k)
                    for c in v["Columns"].keys():
                        t.Columns[c] = Column(eval(v["Columns"][c]["Type"]),v["Columns"][c]["Options"],self.Save)
                        t.Columns[c].Data = v["Columns"][c]["Data"]
            self.autosave = wasOnAutoSave
        except Exception as e:
            raise SystemExit("Error while Loading Database : Incorrect Password" if self.password != "" else "Error while Loading Database : Database is Locked - Please Make Sure You Provide a Password :\n Database(Path,YOUR_PASSWORD_HERE)")
    
    
    def LoadFromSQL(self,SQLFilePath):
        SQL_TYPE_MAP = {
            "INTEGER":int,
            "REAL":float,
            "TEXT":str,
            "BLOB":bytes
        }
        conn = sqlite3.connect(SQLFilePath)
        c = conn.cursor()
        c.execute("SELECT name FROM sqlite_master WHERE type='table';")
        tables = [x[0] for x in c.fetchall() if x[0] != "sqlite_sequence"]
        for table in tables:
            t = self.CreateTable(table)
            c.execute(f"PRAGMA table_info({table})")
            columns = {x[1]:SQL_TYPE_MAP[x[2]] for x in c.fetchall()}
            t.AddColumn(**columns)
            for column in columns:
                t.Columns[column].Data = [x[0] for x in c.execute(f"SELECT {column} FROM {table}").fetchall()]
        conn.close()
    
    def Save(self,Path="",Password="",asJson=False):
        if self.path == "" and Path == "":
            raise SystemExit("Error while Saving Database : No path provided")
        
        if self.path == "":
            self.path = Path
        if self.password == "":
            self.password = Password
        
        file = json.dumps(self.toJson())
        with open(self.path,"w") as f:
            if asJson:
                f.write(file)
            else:
                f.write(encrypt(file,self.password))
        
    def ChangePassword(self,Password):
        self.password = Password

    def SaveIfPossible(self):
        if self.autosave: self.Save()
            
    def toJson(self):
        Tables = {k:v.toJson() for k,v in self.Tables.items()}
        return {"Tables":Tables }
            
    def __str__(self):
        indexes = iter(range(0, len(self.GetTables())))
        tables_list = "\n ".join([f"{next(indexes)} - {table}" for table in self.GetTables()])
        print(f"Database '{self.path}' Contains {len(self)} Tables :\n {tables_list}")
        TableToPrint = input("Select Table To Display (use index or name) : ")
        try :
            index = int(TableToPrint)
            Table = self.GetTables()[index]
            print(self.Tables[Table])
        except :
            print(self.Tables[TableToPrint])
            
        return "\b"


from typing import Type, Callable, Optional
import json
import re

class Column:
    def __init__(self, ctype: Type, options=None, autosave_callback = None) -> None:
        if options is None:
            options = []
        self.Type: Type = ctype
        self.Data: list = []
        self.Options: list[str] = options
        self.Save = autosave_callback
            

    def __gt__(self, other) -> list:
        return [value > other for value in self.Data]

    def __lt__(self, other) -> list:
        return [value < other for value in self.Data]

    def __eq__(self, other) -> list:
        return [value == other for value in self.Data]

    def __ge__(self, other) -> list:
        return [value >= other for value in self.Data]

    def __le__(self, other) -> list:
        return [value <= other for value in self.Data]

    def __ne__(self, other) -> list:
        return [value != other for value in self.Data]

    def between(self, min_val, max_val) -> list:
        return [min_val < value < max_val for value in self.Data]

    def match(self, regex) -> list:
        return [re.match(regex, value) is not None for value in self.Data]

    def __len__(self) -> int:
        return len(self.Data)

    def Add(self, *values) -> None:
        for value in values:
            if not isinstance(value, self.Type) and value is not None:
                raise ValueError(f"Value {value} is not of type {self.Type}")
            self.Data.append(value)
        if self.Save != None : self.Save()

    def RemoveFirst(self) -> None:
        self.Data.pop(0)
        if self.Save != None : self.Save()

    def RemoveLast(self) -> None:
        self.Data.pop()
        if self.Save != None : self.Save()

    def RemoveAll(self, value) -> None:
        self.Data = [i for i in self.Data if i != value]
        if self.Save != None : self.Save()

    def RemoveAt(self, index) -> None:
        self.Data.pop(index)
        if self.Save != None : self.Save()

    def RemoveIf(self, func) -> None:
        self.Data = [i for i in self.Data if not func(i)]
        if self.Save != None : self.Save()

    def RemoveByList(self, values) -> None:
        self.Data = [self.Data[i] for i in range(len(self.Data)) if not values[i]]
        if self.Save != None : self.Save()

    def GetIf(self, func) -> list:
        return [i for i in self.Data if func(i)]

    def Get(self, index) -> Type:
        return self.Data[index]

    def Apply(self, func) -> None:
        self.Data = list(map(func, self.Data))
        if self.Save != None : self.Save()

    def ApplyIf(self, func, condition) -> None:
        self.Data = [func(i) if condition(i) else i for i in self.Data]
        if self.Save != None : self.Save()

    def ReType(self, newType: Type) -> None:
        try:
            self.Data = [newType(i) for i in self.Data]
            self.Type = newType
            if self.Save != None : self.Save()
        except Exception as e:
            raise ValueError(f"Cannot convert to {newType}: {e}")

    def isEmpty(self) -> bool:
        return len(self.Data) == 0

    def __str__(self) -> str:
        data = ""
        max_len = len(str(len(self.Data)))
        for i in range(len(self.Data)):
            data += f"{str(i).ljust(max_len)} | {self.Data[i]}\n"
        return f"Column {self.Type.__name__} with {len(self)} elements \n{data[:-1] if data else 'Empty'}"

    def toJson(self):
        return {"Type":self.Type.__name__,"Data":self.Data,"Options":self.Options}


import json

from regex import R
from .Column import Column
from typing import Optional, Union, Type, Callable

class Table:
    PrintPadding = 1
    def __init__(self, TableName,SaveCallback=None):
        self.Columns: dict[str, Column] = {}
        self.TableName = TableName
        self.Save = SaveCallback
        
        # Events
        self.onInsert    = None
        self.onUpdate    = None
        self.onDelete    = None
        self.onRename    = None
        self.onSelect    = None
        self.onSort      = None
        self.onLimit     = None
        self.onAddColumn = None
        self.onCopy      = None
        
    def __getitem__(self, Key) -> Column:
        return self.Columns[Key]

    def __getattr__(self, name: str) -> Optional[Column]:
        if name in self.Columns:
            return self.Columns[name]
        raise AttributeError(f"Column '{name}' not found.")

    def AddColumn(self, **columns: Union[list, Type]):
        for ColumnName, ColumnData in columns.items():
            if isinstance(ColumnData, list):
                self.Columns[ColumnName] = Column(ColumnData[0], ColumnData[1:],self.Save)
            elif isinstance(ColumnData, type):
                self.Columns[ColumnName] = Column(ColumnData, [],self.Save)
            setattr(self.__class__, ColumnName, property(lambda self: self.Tables[ColumnName]))
        if self.Save != None : self.Save()
        if self.onAddColumn != None: self.onAddColumn(self)

    def RenameColumn(self, OldName, NewName):
        self.Columns[NewName] = self.Columns.pop(OldName)
        setattr(self.__class__, NewName, property(lambda self: self.Columns[NewName]))
        delattr(self.__class__, OldName)
        if self.Save != None : self.Save()
        if self.onRename != None: self.onRename(self)
        
    def RemoveColumn(self, ColumnName):
        del self.Columns[ColumnName]
        delattr(self.__class__, ColumnName)
        if self.Save != None : self.Save()
        if self.onRename != None: self.onRename(self)
        
        
        
    def Insert(self, **columns: Union[list, Type]):
        for k,v in self.Columns.items():
            if k in list(columns.keys()):
                v.Add(v.Type(columns[k]))
            else:
                v.Add(v.Type())
        if self.Save != None : self.Save()
        if self.onInsert != None: self.onInsert(self)

    # @NotImplemented    
    # def InsertObject(self,obj):
    #     print(obj.__dict__)
    #     try:
    #         for k,v in self.Columns.items():
    #             if k in obj.__dict__.keys():
    #                 v.Add(v.Type(obj.__dict__[k]))
    #             else:
    #                 v.Add(v.Type())
    #     except Exception as e:
    #         print(f"Error: {e}")    
    #     if self.Save != None : self.Save()

            
                
    
    def Select(self,condition=None) -> "Table":
        if condition == None:
            ret = self.Copy()
            if self.onSelect != None: self.onSelect(ret)
            return ret
        ReturnTable = Table("Selected Table")
        ReturnTable.Columns = {k: Column(v.Type, v.Options) for k,v in self.Columns.items()}
        for i in range(self.RowCount):
            if condition[i]:
                ReturnTable.Insert(**{k:self.Rows[i][j] for j,k in enumerate(self.Columns.keys())})
        if self.onSelect != None: self.onSelect(ReturnTable)
        return ReturnTable

    def Delete(self,index=None,where=None):
        if self.isEmpty():raise ValueError("Table is empty.")
        if index == None and where == None:
            self.Columns = {k: Column(v.Type, v.Options) for k,v in self.Columns.items()}
        if index != None:
            [v.RemoveAt(index) for v in self.Columns.values()]
        if where != None:
            [x.RemoveByList(where) for x in self.Columns.values()]
        if self.Save != None : self.Save()
        if self.onDelete != None: self.onDelete(self)
        

    def Update(self, index=None, where=None, **columns: Union[list, Type]):
        if index == None and where == None:
            for k,v in columns.items():
                self.Columns[k].Data = [v for _ in self.Columns[k].Data]
        if index != None:
            for k,v in columns.items():
                self.Columns[k].Data[index] = v
        if where != None:
            for k,v in columns.items():
                for i in range(len(where)):
                    if where[i]:
                        self.Columns[k].Data[i] = v
        if self.Save != None : self.Save()
        if self.onUpdate != None: self.onUpdate(self)

        

    def isEmpty(self) -> bool:
        return all([col.isEmpty() for col in self.Columns.values()])
    
    def __str__(self) -> str:
        if self.isEmpty():return f"Table is empty."
        def getMaxLength(column) -> int:return max([len(str(x)) for x in column.Data])
        def formatCell(cell, maxLen):return f" {cell} {' '*(maxLen-len(str(cell)))}"
        maxLenperColumn = [max(getMaxLength(col), len(key))+self.PrintPadding for col, key in zip(self.Columns.values(), self.Columns.keys())]
        numberOfLines = max([len(col.Data) for col in self.Columns.values()])
        ret = f"{self.TableName}:\n"
        ret += "|"+"|".join([formatCell(list(self.Columns.keys())[i], maxLenperColumn[i]) for i in range(len(self.Columns))]) + "|\n"
        ret +=  "|"+"|\n|".join("|".join((formatCell(col.Data[i], maxLen) if col.Type!=bytes else "BLOB") for col, maxLen in zip(self.Columns.values(), maxLenperColumn)) for i in range(numberOfLines)) + "|"
        return ret
    
    @property
    def Rows(self) -> list:
        return [[col.Data[i] for col in self.Columns.values()] for i in range(len(self.Columns[list(self.Columns.keys())[0]].Data))]
    
    def GetColumns(self) -> list:
        return list(self.Columns.keys())
    
    def Exists(self, **columns: Union[list, Type]) -> bool:
        for k,v in columns.items():
            if k not in self.Columns.keys():return False
            if v not in self.Columns[k].Data:return False
        return True
    
    @property
    def RowCount(self) -> int:
        return len(self.Rows)
    
    def Limit(self,n):
        T = Table(self.TableName,self.Save)
        T.Columns = {k: Column(v.Type, v.Options) for k,v in self.Columns.items()}
        for i in range(min(n,self.RowCount)):
            T.Insert(**{k:v.Data[i] for k,v in self.Columns.items()})
        if self.onLimit != None: self.onLimit(T)
        return T
    
    def Sort(self, Column, Reverse=False):
        if Column not in self.Columns.keys():
            raise ValueError(f"Column '{Column}' not found.")
        T = self.Copy()
        sorted_indices = sorted(range(len(T.Columns[Column].Data)), key=lambda i: T.Columns[Column].Data[i], reverse=Reverse)
        for col in T.Columns.values():
            col.Data = [col.Data[i] for i in sorted_indices]
        if self.onSort != None: self.onSort(T)
        return T
    
    def Copy(self):
        T = Table(self.TableName,self.Save)
        T.Columns = {k: Column(v.Type, v.Options) for k,v in self.Columns.items()}
        for i in range(self.RowCount):
            T.Insert(**{k:v.Data[i] for k,v in self.Columns.items()})
        if self.onCopy != None: self.onCopy(T)
        return T
    
    def toJson(self):
        return {"Columns":{k:v.toJson() for k,v in self.Columns.items()}}